# $Id: pybdblp.py,v$
#
# Author: Jaap-Henk Hoepman (jhh@xs4all.nl)
#
# Functions to retrieve bibtex entries from the DBLP repository.
#
# Bibtex Entries in DBLP are html files that look like this
#
# [.. prefix with html header]
# <pre>@...{<a href=...>DBLP</a>:key
# </pre>
# <pre>..
# </pre>
# [.. postfix with html footer]


import sys, re, urllib.request, pybscan, pybparse


# DBLP prefix (put before citation key to get location to retrieve
# the bibtex entry from)
dblp_url_prefix = "http://dblp.uni-trier.de/rec/bibtex/"

# Regular expression matching DBLP keys (keys not matching are not even tried
# to be retrieved from the DBLP server to avoid overloading it with requests
dblp_key_re = re.compile('DBLP.*')

# Regular expression matching the DBLP href link
# <a href=...>DBLP</a>:key
# present in a key
# of a DBLP bibtex entry as retrieved from the DBLP website
dblp_href_re = re.compile('<[^>]*>DBLP</a>')

# Remove the DBLP href link present in a key of a DBLP bibtex entry
# and replace it with the string DBLP
def strip_link ( entry ):
    return re.sub(dblp_href_re,'DBLP',entry)

# Retrieve the citation 'key' from the DBLP repository.
# keys must match dblp_key_re or else they are not retrieved
# Returns the citation (possibly including cross referenced entry) as a 
# string. Returns an empty string on any error (eg: no network)
def dblp_get(key):
    try:
        # avoid overloading DBLP with unneccesary questions
        if not dblp_key_re.match(key):
            return ''
        # strip the DBLP: part from key and prefix with DBLP url prefix
        # and open the html page at that url as a file
        f = urllib.request.urlopen(dblp_url_prefix + key[5:])
        # read the file as one single stream of bytes (!),
        # and turn that into a string
        # FIXME: JHH: this uses a default encoding!!
        s = (f.read()).decode()
        # split the string at each occurance of <pre>
        entries = re.split('<pre>',s)
        # and process entries[1] (and entries[2] if present)
        # split each at </pre> and keep the first part
        entry = re.split('</pre>',entries[1])
        ret = strip_link(entry[0])    
        if len(entries) > 2:
            entry = re.split('</pre>',entries[2])
            ret = ret + '\n' + strip_link(entry[0])    
        return ret
    except:
        return ''

# Try to obtain the bibtex entry for the key from DBLP
# and return a BibEntry if succesful. Otherwise return None
def get_entry_if_possible(key):
    # get the entry from DBLP as a string
    s = dblp_get(key)
    if s != '':
        entries = pybparse.parse_string(s,None)
        return entries[key]
    else:
        return None


