# $Id: pybparse.py,v 1.1 2009/08/19 08:05:04 jhh Exp $
#
# Author: Jaap-Henk Hoepman (jhh@xs4all.nl)
#
# Functions to parse BiBTeX files
#
# - A bibtex record is stored as python dictionary, where the key in the
#   dictionary (a string with double quotes) corresponds to the bibtex
#   fieldname (without those quotes), and the value stores
#   the value (in both cases with "" or {}). The dictionary holds
#   an extra key "mytype" for the type of the bibtex entry
#   (but stored without quotes).
#
# - All records in a bibtex file are returned as a dictionary of records
#   (ie a dictionary of dictionaries), where the key in the dictionary
#   for a record in the dictionary corresponds to its cite key
#
# - The first lines in the bibtex file, upto the first @ at the start of a
#   new line are ignored (because bibtex does that too, traditionally)
#   Comments (starting with % and ending at a newline) are also skipped
#
#   Actually, between entries, anyline that doesn't start with a @ is ignored
#
# - We use the following simplified syntax for a bibtex record/entry
#   (whitespace/comments allowed between tokens)
#   string = "@" <stringsym> "{" <char>* "}"
#   stringsym = "string" | "STRING"
#   entry = "@" <name> "{" <key> "," <field> ("," <field>)+ "}"
#   name = <namechars>+
#   key = <keychars>+
#   field = <fieldname> "=" <value>
#   value = any sequence of characters terminated by either a "," or a "}"
#           that does not occur inside a matching pairs of "{" "}" or """ """
#           (where, in the first case, the "}" would close the nesting, and
#           where "\" preceding these chars ensures they are ignored regarding nesting


# TODO
#
# parser trips on certain files, without giving an error message
# eg records ending with ... }# bla
# prove to be a problem
#
# This should really be auto-generated from a proper syntax...
#

# adapted to python3

import string, sys
from pybbib import BibEntry
from pybscan import BibScanner

# the characters that may appear in a
# citation key; this is NOT COMPLETE
# 9-12-2008: added / to key_letters because DBLP uses it
key_letters = string.ascii_letters + string.digits + "?_-+:;*.()/"

# the characters that may appear in a fieldname, the record type
# this is NOT COMPLETE
field_letters = string.ascii_letters + string.digits + "?_-+:;*."

# skip leading whitespace and read a name/key, and return it. curch()
# points to the first non-name char. if at EOF, return None
def bib_getname(bs,allowed_letters):
    ret = bs.expect(allowed_letters)
    if ret == None:
        return None
    name = ret
    while bs.curch() in allowed_letters:
        name = name + bs.curch()
        bs.nextch()
    return name

# Get the value for a bibtex field. Instead of parsing the field, we use
# the following rule: a field is terminated by either a "," or a "}"
# (which is not following a "\"), unless either of them occurs within a quoted
# context (ie inside an unclosed { or ")
#
# Note: this fails to detect errors like "field = {test}} # jan,
#   where the 2nd } is erroneous
#
# Return the value as a string. curch() points to the first char not part
# of the value (ie "," or "}")
#
# 9-12-2008: This routine strips a trailing \n from the value, if for instance
# the value is terminated by a } on a new line (which happens for
# the last field in an entry that doesn't have to be terminated by a ,
def bib_getvalue(bs):
    s = ""
    symstack = []
    while symstack != [] or ( bs.curch() not in [",", "}"] ) :
        s = s + bs.curch()
        if bs.prevch() != "\\":
            if bs.curch() == "{":
                symstack.append("{")
            elif bs.curch() == "\"":
                if (symstack != []) and (symstack[-1] == "\""):
                    symstack.pop()
                else:
                    symstack.append("\"")
            elif bs.curch() == "}" and symstack != [] and symstack[-1] == "{":
                symstack.pop()
        bs.nextch()
    if (bs.prevch() == '\n'):
        return s[:-1] 
    else:
        return s
    
# return the next (fieldname,value) pair, or None if not at a fieldname
def bib_getfield(bs):
    fieldname = bib_getname(bs,field_letters)
    if fieldname == None:
        return None
    bs.required("=")
    bs.skip_whitespace()
    value = bib_getvalue(bs)
    bs.expect(",")
    return (fieldname, value)

# return the next entry as a BibEntry or None if no more entries
# returns a tuple (key,entry as dictionary)
# key is "<string>" if the entry is a string
def bib_getentry(bs):
    values = {}
    bs.skip_until_at_line()
    if bs.expect("@") == None:
        return None
    tp = bib_getname(bs,field_letters) # TODO: this may return None!
    if (tp == None):
        bs.error("Expected a field type")
    bs.required("{")
    # strings must be ignored and do not have a key
    if tp != "string":
        key = bib_getname(bs,key_letters)
        bs.required(",")
    else:
        key = "<string>"
    nextfield = bib_getfield(bs)
    while nextfield != None:
        (fieldname,value) = nextfield
        values[fieldname] = value
        nextfield = bib_getfield(bs)
    bs.required("}")
    be = BibEntry(tp,key,values)
    return (key,be)


# Scan  and return a dictionary with for each key a dictionary
# of all the fields for that record; Throws excpetion if file does not
# exist or error occured
#
# Calls progress_callback for each succesfully loaded record, unless it
# equals None
def parse_with_scanner(bs,progress_callback):
    entries = {} 
    nextentry = bib_getentry(bs)
    while nextentry != None:
        (key,entry) = nextentry
        if progress_callback != None:
            progress_callback(key,entry)
        if key != "<string>":
            entries[key] = entry
        nextentry = bib_getentry(bs)
    return entries

# Load bib file fname and return a dictionary with for each key a dictionary
# of all the fields for the corresponding record; Throws excpetion if file does not
# exist or error occured
#
# Calls progress_callback for each succesfully loaded record.
def parse_file(fname,progress_callback):
    bs = BibScanner(fname,True)
    return parse_with_scanner(bs,progress_callback)

# Load all bibtex records contained in string s
# Return a dictionary with for each cite key a dictionary
# of all the fields for the corresponding bibtex record;
# Throws excpetion if an error occured
def parse_string(s,progress_callback):
    bs = BibScanner(s,False)
    return parse_with_scanner(bs,progress_callback)

